const char *help = "\
MLPTorch III (c) Trebolloc & Co 2003\n\
\n\
This program will train a MLP with log-softmax outputs\n\
for classification and linear outputs for regression.\n\
For classification, the criterion will be a cross-entropy criterion.\n\
For regression, it uses a mean-squared error criterion.\n";

#include "torch/MatDataSet.h"
#include "torch/ClassFormatDataSet.h"
#include "torch/ClassNLLCriterion.h"
#include "torch/MSECriterion.h"
#include "torch/OneHotClassFormat.h"
#include "torch/ClassMeasurer.h"
#include "torch/MSEMeasurer.h"

#include "torch/StochasticGradient.h"
#include "torch/KFold.h"

#include "torch/ConnectedMachine.h"
#include "torch/Linear.h"
#include "torch/Tanh.h"
#include "torch/LogSoftMax.h"

#include "torch/MeanVarNorm.h"
#include "torch/DiskXFile.h"
#include "torch/CmdLine.h"
#include "torch/Random.h"

using namespace Torch;

int main(int argc, char **argv)
{
  char *valid_file = "";
  char *file;

  int n_inputs = 2;
  int n_targets = 1;
  int n_hu = 1;

  int max_load;
  int max_load_valid;

  char *model_file;
  int k_fold;
  bool binary_mode = false;
  bool regression_mode;
  int n_outputs;

  CmdLine cmd;

  
  int mode = cmd.read(argc, argv);

  Allocator *allocator = new Allocator;
  DiskXFile::setLittleEndianMode();


  //=================== Create the MLP... =========================
  n_inputs = 3;
  ConnectedMachine mlp;

  Linear *c1 = new(allocator) Linear(n_inputs, n_hu);
  Tanh *c2 = new(allocator) Tanh(n_hu);
  Linear *c3 = new(allocator) Linear(n_hu, n_outputs);

  mlp.addFCL(c1);    
  mlp.addFCL(c2);
  mlp.addFCL(c3);

  // Initialize the MLP
  mlp.build();
  mlp.setPartialBackprop();

  //=================== DataSets & Measurers... ===================

  // Create the training dataset (normalize inputs)
  DataSet *data = NULL;
  MeanVarNorm *mv_norm = NULL;
  
  data = new(allocator) MatDataSet(file, n_inputs, n_targets, false, max_load, binary_mode);
  mv_norm = new(allocator) MeanVarNorm(data);
  data->preProcess(mv_norm);

  // The list of measurers...
  MeasurerList measurers;

  // The class format
  OneHotClassFormat *class_format = NULL;

  // The validation set...
  if(mode == 0)
  {
    // Create a validation set, if any
    if(strcmp(valid_file, ""))
    {
      // Load the validation set and normalize it with the
      // values in the train dataset
      DataSet *valid_data = NULL;
      MatDataSet *valid_mat_data = new(allocator) MatDataSet(valid_file, n_inputs, 1, false, max_load_valid, binary_mode);
      valid_mat_data->preProcess(mv_norm);

      valid_data = new(allocator) ClassFormatDataSet(valid_mat_data, n_targets);
      
      ClassMeasurer *valid_class_meas = new(allocator) ClassMeasurer(mlp.outputs, valid_data, class_format, cmd.getXFile("the_valid_class_err"));
      measurers.addNode(valid_class_meas);
    }
  }
  
  // Measurers on the training dataset
  ClassMeasurer *class_meas = new(allocator) ClassMeasurer(mlp.outputs, data, class_format, cmd.getXFile("the_class_err"));
  measurers.addNode(class_meas);
  
  //=================== The Trainer ===============================
  
  // The criterion for the StochasticGradient (MSE criterion or NLL criterion)
  Criterion *criterion = NULL;
  if(regression_mode)
    criterion = new(allocator) MSECriterion(n_outputs);
  else
    criterion = new(allocator) ClassNLLCriterion(class_format);

  // The Gradient Machine Trainer
  StochasticGradient trainer(&mlp, criterion);

  //=================== Let's go... ===============================

  // Print the number of parameter of the MLP (just for fun)
  message("Number of parameters: %d", mlp.params->n_params);

  if(mode == 0)
  {
    if(k_fold <= 0)
    {
      trainer.train(data, &measurers);
      
      if(strcmp(model_file, ""))
      {
        DiskXFile model_(model_file, "w");
        cmd.saveXFile(&model_);
        mv_norm->saveXFile(&model_);
        mlp.saveXFile(&model_);
      }
    }
    else
    {
      KFold k(&trainer, k_fold);
      k.crossValidate(data, NULL, &measurers);
    }
  }
  else
    trainer.test(&measurers);

  delete allocator;

  return(0);
}
